@@ -22,7 +22,6 @@ from oslo_log import log as logging
 from oslo_utils import excutils
 import six
 
-from neutron._i18n import _LE, _LW
 from neutron.agent.l3 import dvr_fip_ns
 from neutron.agent.l3 import dvr_router_base
 from neutron.agent.linux import ip_lib
@@ -39,28 +38,37 @@ Arp_entry = collections.namedtuple(
 
 
 class DvrLocalRouter(dvr_router_base.DvrRouterBase):
-    def __init__(self, agent, host, *args, **kwargs):
-        super(DvrLocalRouter, self).__init__(agent, host, *args, **kwargs)
+    def __init__(self, host, *args, **kwargs):
+        super(DvrLocalRouter, self).__init__(host, *args, **kwargs)
 
         self.floating_ips_dict = {}
+        self.centralized_floatingips_set = set()
         # Linklocal subnet for router and floating IP namespace link
         self.rtr_fip_subnet = None
-        self.dist_fip_count = None
+        self.rtr_fip_connect = False
         self.fip_ns = None
         self._pending_arp_set = set()
 
-    def get_floating_ips(self):
-        """Filter Floating IPs to be hosted on this agent."""
-        floating_ips = super(DvrLocalRouter, self).get_floating_ips()
-        return [i for i in floating_ips if (
-                   (i['host'] == self.host) or
-                   (i.get('dest_host') == self.host))]
+    def get_centralized_router_cidrs(self):
+        return self.centralized_floatingips_set
 
-    def floating_forward_rules(self, floating_ip, fixed_ip):
+    def migrate_centralized_floating_ip(self, fip, interface_name, device):
+        # Remove the centralized fip first and then add fip to the host
+        ip_cidr = common_utils.ip_to_cidr(fip['floating_ip_address'])
+        self.floating_ip_removed_dist(ip_cidr)
+        # Now add the floating_ip to the current host
+        self.floating_ip_added_dist(fip, ip_cidr)
+
+    def floating_forward_rules(self, fip):
         """Override this function defined in router_info for dvr routers."""
         if not self.fip_ns:
             return []
 
+        if fip.get(n_const.DVR_SNAT_BOUND):
+            return []
+
+        fixed_ip = fip['fixed_ip_address']
+        floating_ip = fip['floating_ip_address']
         rtr_2_fip_name = self.fip_ns.get_rtr_ext_device_name(self.router_id)
         dnat_from_floatingip_to_fixedip = (
             'PREROUTING', '-d %s/32 -i %s -j DNAT --to-destination %s' % (
@@ -83,8 +91,35 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
             'FORWARD', '-s %s/32 -j $float-snat' % fixed_ip)
         return [mark_traffic_to_floating_ip, mark_traffic_from_fixed_ip]
 
+    def add_centralized_floatingip(self, fip, fip_cidr):
+        """Implements floatingip in centralized network node.
+        This is a dummy function and is overridden in dvr_edge_router.py
+        to add the floatingip function to the snat namespace.
+        """
+
+    def remove_centralized_floatingip(self, fip_cidr):
+        """Removes floatingip from centralized network node.
+        This is a dummy function and is overridden in dvr_edge_router.py
+        to remove the floatingip function from the snat namespace.
+        """
+
     def floating_ip_added_dist(self, fip, fip_cidr):
-        """Add floating IP to FIP namespace."""
+        """Add floating IP to respective namespace based on agent mode."""
+        if fip.get(n_const.DVR_SNAT_BOUND):
+            floating_ip_status = self.add_centralized_floatingip(fip, fip_cidr)
+            if floating_ip_status == lib_constants.FLOATINGIP_STATUS_ACTIVE:
+                self.centralized_floatingips_set.add(fip_cidr)
+            return floating_ip_status
+        if not self._check_if_floatingip_bound_to_host(fip):
+            # TODO(Swami): Need to figure out what status
+            # should be returned when the floating IP is
+            # not destined for this agent and if the floating
+            # IP is configured in a different compute host.
+            # This should not happen once we fix the server
+            # side code, but still a check to make sure if
+            # the floating IP is intended for this host should
+            # be done.
+            return
         floating_ip = fip['floating_ip_address']
         fixed_ip = fip['fixed_ip_address']
         self._add_floating_ip_rule(floating_ip, fixed_ip)
@@ -102,14 +137,12 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
                 self.fip_ns.agent_gateway_port['id']))
         ip_lib.send_ip_addr_adv_notif(fip_ns_name,
                                       interface_name,
-                                      floating_ip,
-                                      self.agent_conf)
-        # update internal structures
-        self.dist_fip_count = self.dist_fip_count + 1
+                                      floating_ip)
+        return lib_constants.FLOATINGIP_STATUS_ACTIVE
 
     def _add_floating_ip_rule(self, floating_ip, fixed_ip):
         rule_pr = self.fip_ns.allocate_rule_priority(floating_ip)
-        self.floating_ips_dict[floating_ip] = rule_pr
+        self.floating_ips_dict[floating_ip] = (fixed_ip, rule_pr)
         ip_rule = ip_lib.IPRule(namespace=self.ns_name)
         ip_rule.rule.add(ip=fixed_ip,
                          table=dvr_fip_ns.FIP_RT_TBL,
@@ -117,9 +150,9 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
 
     def _remove_floating_ip_rule(self, floating_ip):
         if floating_ip in self.floating_ips_dict:
-            rule_pr = self.floating_ips_dict[floating_ip]
+            fixed_ip, rule_pr = self.floating_ips_dict[floating_ip]
             ip_rule = ip_lib.IPRule(namespace=self.ns_name)
-            ip_rule.rule.delete(ip=floating_ip,
+            ip_rule.rule.delete(ip=fixed_ip,
                                 table=dvr_fip_ns.FIP_RT_TBL,
                                 priority=rule_pr)
             self.fip_ns.deallocate_rule_priority(floating_ip)
@@ -127,31 +160,23 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
 
     def floating_ip_removed_dist(self, fip_cidr):
         """Remove floating IP from FIP namespace."""
+        if fip_cidr in self.centralized_floatingips_set:
+            self.remove_centralized_floatingip(fip_cidr)
+            self.centralized_floatingips_set.remove(fip_cidr)
+            return
         floating_ip = fip_cidr.split('/')[0]
-        rtr_2_fip_name = self.fip_ns.get_rtr_ext_device_name(self.router_id)
         fip_2_rtr_name = self.fip_ns.get_int_device_name(self.router_id)
         if self.rtr_fip_subnet is None:
-            self.rtr_fip_subnet = self.fip_ns.local_subnets.allocate(
+            self.rtr_fip_subnet = self.fip_ns.local_subnets.lookup(
                 self.router_id)
+        if self.rtr_fip_subnet:
+            rtr_2_fip, fip_2_rtr = self.rtr_fip_subnet.get_pair()
+            fip_ns_name = self.fip_ns.get_name()
+            self._remove_floating_ip_rule(floating_ip)
 
-        rtr_2_fip, fip_2_rtr = self.rtr_fip_subnet.get_pair()
-        fip_ns_name = self.fip_ns.get_name()
-        self._remove_floating_ip_rule(floating_ip)
+            device = ip_lib.IPDevice(fip_2_rtr_name, namespace=fip_ns_name)
 
-        device = ip_lib.IPDevice(fip_2_rtr_name, namespace=fip_ns_name)
-
-        device.route.delete_route(fip_cidr, str(rtr_2_fip.ip))
-        # check if this is the last FIP for this router
-        self.dist_fip_count = self.dist_fip_count - 1
-        if self.dist_fip_count == 0:
-            #remove default route entry
-            device = ip_lib.IPDevice(rtr_2_fip_name, namespace=self.ns_name)
-            ns_ip = ip_lib.IPWrapper(namespace=fip_ns_name)
-            device.route.delete_gateway(str(fip_2_rtr.ip),
-                                        table=dvr_fip_ns.FIP_RT_TBL)
-            self.fip_ns.local_subnets.release(self.router_id)
-            self.rtr_fip_subnet = None
-            ns_ip.del_veth(fip_2_rtr_name)
+            device.route.delete_route(fip_cidr, str(rtr_2_fip.ip))
 
     def floating_ip_moved_dist(self, fip):
         """Handle floating IP move between fixed IPs."""
@@ -162,8 +187,7 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
     def add_floating_ip(self, fip, interface_name, device):
         # Special Handling for DVR - update FIP namespace
         ip_cidr = common_utils.ip_to_cidr(fip['floating_ip_address'])
-        self.floating_ip_added_dist(fip, ip_cidr)
-        return lib_constants.FLOATINGIP_STATUS_ACTIVE
+        return self.floating_ip_added_dist(fip, ip_cidr)
 
     def remove_floating_ip(self, device, ip_cidr):
         self.floating_ip_removed_dist(ip_cidr)
@@ -215,7 +239,8 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
                 arp_delete.add(arp_entry)
         self._pending_arp_set -= arp_delete
 
-    def _update_arp_entry(self, ip, mac, subnet_id, operation):
+    def _update_arp_entry(
+        self, ip, mac, subnet_id, operation, nud_state='permanent'):
         """Add or delete arp entry into router namespace for the subnet."""
         port = self._get_internal_port(subnet_id)
         # update arp entry only if the subnet is attached to the router
@@ -228,36 +253,41 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
             device = ip_lib.IPDevice(interface_name, namespace=self.ns_name)
             if device.exists():
                 if operation == 'add':
-                    device.neigh.add(ip, mac)
+                    device.neigh.add(ip, mac, nud_state=nud_state)
                 elif operation == 'delete':
                     device.neigh.delete(ip, mac)
                 return True
             else:
                 if operation == 'add':
-                    LOG.warning(_LW("Device %s does not exist so ARP entry "
-                                    "cannot be updated, will cache "
-                                    "information to be applied later "
-                                    "when the device exists"),
+                    LOG.warning("Device %s does not exist so ARP entry "
+                                "cannot be updated, will cache "
+                                "information to be applied later "
+                                "when the device exists",
                                 device)
                     self._cache_arp_entry(ip, mac, subnet_id, operation)
                 return False
         except Exception:
             with excutils.save_and_reraise_exception():
-                LOG.exception(_LE("DVR: Failed updating arp entry"))
+                LOG.exception("DVR: Failed updating arp entry")
 
     def _set_subnet_arp_info(self, subnet_id):
         """Set ARP info retrieved from Plugin for existing ports."""
         # TODO(Carl) Can we eliminate the need to make this RPC while
         # processing a router.
         subnet_ports = self.agent.get_ports_by_subnet(subnet_id)
+        ignored_device_owners = (
+            lib_constants.ROUTER_INTERFACE_OWNERS +
+            tuple(common_utils.get_dvr_allowed_address_pair_device_owners()))
 
         for p in subnet_ports:
-            if p['device_owner'] not in lib_constants.ROUTER_INTERFACE_OWNERS:
+            nud_state = 'permanent' if p.get('device_owner') else 'reachable'
+            if p['device_owner'] not in ignored_device_owners:
                 for fixed_ip in p['fixed_ips']:
                     self._update_arp_entry(fixed_ip['ip_address'],
                                            p['mac_address'],
                                            subnet_id,
-                                           'add')
+                                           'add',
+                                           nud_state=nud_state)
         self._process_arp_cache_for_internal_port(subnet_id)
 
     @staticmethod
@@ -320,11 +350,10 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
 
     def _snat_redirect_modify(self, gateway, sn_port, sn_int, is_add):
         """Adds or removes rules and routes for SNAT redirection."""
+        cmd = ['net.ipv4.conf.%s.send_redirects=0' % sn_int]
         try:
             ns_ipr = ip_lib.IPRule(namespace=self.ns_name)
             ns_ipd = ip_lib.IPDevice(sn_int, namespace=self.ns_name)
-            if is_add:
-                ns_ipwrapr = ip_lib.IPWrapper(namespace=self.ns_name)
             for port_fixed_ip in sn_port['fixed_ips']:
                 # Iterate and find the gateway IP address matching
                 # the IP version
@@ -342,9 +371,7 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
                             ns_ipr.rule.add(ip=sn_port_cidr,
                                             table=snat_idx,
                                             priority=snat_idx)
-                            ns_ipwrapr.netns.execute(
-                                ['sysctl', '-w',
-                                 'net.ipv4.conf.%s.send_redirects=0' % sn_int])
+                            ip_lib.sysctl(cmd, namespace=self.ns_name)
                         else:
                             self._delete_gateway_device_if_exists(ns_ipd,
                                                                   gw_ip_addr,
@@ -354,10 +381,10 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
                                                priority=snat_idx)
         except Exception:
             if is_add:
-                exc = _LE('DVR: error adding redirection logic')
+                exc = 'DVR: error adding redirection logic'
             else:
-                exc = _LE('DVR: snat remove failed to clear the rule '
-                          'and device')
+                exc = ('DVR: snat remove failed to clear the rule '
+                       'and device')
             LOG.exception(exc)
 
     def _snat_redirect_add(self, gateway, sn_port, sn_int):
@@ -376,15 +403,24 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
         # entries for the dvr services ports into the router
         # namespace. This does not have dependency on the
         # external_gateway port or the agent_mode.
+        ex_gw_port = self.get_ex_gw_port()
         for subnet in port['subnets']:
             self._set_subnet_arp_info(subnet['id'])
+            if ex_gw_port:
+                # Check for address_scopes here if gateway exists.
+                if self._check_if_address_scopes_match(port, ex_gw_port):
+                    self._add_interface_routing_rule_to_router_ns(port)
+                    self._add_interface_route_to_fip_ns(port)
         self._snat_redirect_add_from_port(port)
 
     def _snat_redirect_add_from_port(self, port):
         ex_gw_port = self.get_ex_gw_port()
         if not ex_gw_port:
             return
-
+        if self._check_if_address_scopes_match(port, ex_gw_port):
+            # If address scopes match there is no need to cleanup the
+            # snat redirect rules, hence return here.
+            return
         sn_port = self.get_snat_port_for_internal_port(port)
         if not sn_port:
             return
@@ -393,9 +429,21 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
         self._snat_redirect_add(sn_port, port, interface_name)
 
     def _dvr_internal_network_removed(self, port):
+        # Clean up the cached arp entries related to the port subnet
+        for subnet in port['subnets']:
+            self._delete_arp_cache_for_internal_port(subnet)
+
         if not self.ex_gw_port:
             return
 
+        # Delete DVR address_scope static route for the removed interface
+        # Check for address_scopes here.
+        if self._check_if_address_scopes_match(port, self.ex_gw_port):
+            self._delete_interface_route_in_fip_ns(port)
+            self._delete_interface_routing_rule_in_router_ns(port)
+            # If address scopes match there is no need to cleanup the
+            # snat redirect rules, hence return here.
+            return
         sn_port = self.get_snat_port_for_internal_port(port, self.snat_ports)
         if not sn_port:
             return
@@ -403,9 +451,6 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
         # DVR handling code for SNAT
         interface_name = self.get_internal_device_name(port['id'])
         self._snat_redirect_remove(sn_port, port, interface_name)
-        # Clean up the cached arp entries related to the port subnet
-        for subnet in port['subnets']:
-            self._delete_arp_cache_for_internal_port(subnet)
 
     def internal_network_removed(self, port):
         self._dvr_internal_network_removed(port)
@@ -417,24 +462,41 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
         return next(
             (p for p in fip_ports if p['network_id'] == ext_net_id), None)
 
+    def get_snat_external_device_interface_name(self, port_id):
+        pass
+
     def get_external_device_interface_name(self, ex_gw_port):
         fip_int = self.fip_ns.get_int_device_name(self.router_id)
         if ip_lib.device_exists(fip_int, namespace=self.fip_ns.get_name()):
             return self.fip_ns.get_rtr_ext_device_name(self.router_id)
 
+    def enable_snat_redirect_rules(self, ex_gw_port):
+        for p in self.internal_ports:
+            if not self._check_if_address_scopes_match(p, ex_gw_port):
+                gateway = self.get_snat_port_for_internal_port(p)
+                if not gateway:
+                    continue
+                internal_dev = self.get_internal_device_name(p['id'])
+                self._snat_redirect_add(gateway, p, internal_dev)
+
+    def disable_snat_redirect_rules(self, ex_gw_port):
+        for p in self.internal_ports:
+            if not self._check_if_address_scopes_match(p, ex_gw_port):
+                gateway = self.get_snat_port_for_internal_port(
+                    p, self.snat_ports)
+                if not gateway:
+                    continue
+                internal_dev = self.get_internal_device_name(p['id'])
+                self._snat_redirect_remove(gateway, p, internal_dev)
+
     def external_gateway_added(self, ex_gw_port, interface_name):
         # TODO(Carl) Refactor external_gateway_added/updated/removed to use
         # super class implementation where possible.  Looks like preserve_ips,
         # and ns_name are the key differences.
-        ip_wrapr = ip_lib.IPWrapper(namespace=self.ns_name)
-        ip_wrapr.netns.execute(['sysctl', '-w',
-                               'net.ipv4.conf.all.send_redirects=0'])
-        for p in self.internal_ports:
-            gateway = self.get_snat_port_for_internal_port(p)
-            id_name = self.get_internal_device_name(p['id'])
-            if gateway:
-                self._snat_redirect_add(gateway, p, id_name)
+        cmd = ['net.ipv4.conf.all.send_redirects=0']
+        ip_lib.sysctl(cmd, namespace=self.ns_name)
 
+        self.enable_snat_redirect_rules(ex_gw_port)
         for port in self.get_snat_interfaces():
             for ip in port['fixed_ips']:
                 self._update_arp_entry(ip['ip_address'],
@@ -452,19 +514,15 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
             to_fip_interface_name = (
                 self.get_external_device_interface_name(ex_gw_port))
             self.process_floating_ip_addresses(to_fip_interface_name)
+            # Remove the router to fip namespace connection after the
+            # gateway is removed.
+            self.fip_ns.delete_rtr_2_fip_link(self)
+            self.rtr_fip_connect = False
         # NOTE:_snat_redirect_remove should be only called when the
         # gateway is cleared and should not be called when the gateway
         # is moved or rescheduled.
         if not self.router.get('gw_port'):
-            for p in self.internal_ports:
-                # NOTE: When removing the gateway port, pass in the snat_port
-                # cache along with the current ports.
-                gateway = self.get_snat_port_for_internal_port(
-                    p, self.snat_ports)
-                if not gateway:
-                    continue
-                internal_interface = self.get_internal_device_name(p['id'])
-                self._snat_redirect_remove(gateway, p, internal_interface)
+            self.disable_snat_redirect_rules(ex_gw_port)
 
     def _handle_router_snat_rules(self, ex_gw_port, interface_name):
         """Configures NAT rules for Floating IPs for DVR."""
@@ -509,45 +567,163 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
             ext_scope_mark)
         return ports_scopemark
 
-    def process_external(self, agent):
-        ex_gw_port = self.get_ex_gw_port()
-        if ex_gw_port:
-            self.create_dvr_fip_interfaces(ex_gw_port)
-        super(DvrLocalRouter, self).process_external(agent)
+    def _check_if_floatingip_bound_to_host(self, fip):
+        """Check if the floating IP is bound to this host."""
+        return self.host in (fip.get('host'), fip.get('dest_host'))
 
-    def create_dvr_fip_interfaces(self, ex_gw_port):
-        floating_ips = self.get_floating_ips()
+    def process_external(self):
+        if self.agent_conf.agent_mode != (
+            n_const.L3_AGENT_MODE_DVR_NO_EXTERNAL):
+            ex_gw_port = self.get_ex_gw_port()
+            if ex_gw_port:
+                self.create_dvr_external_gateway_on_agent(ex_gw_port)
+                self.connect_rtr_2_fip()
+        super(DvrLocalRouter, self).process_external()
+
+    def connect_rtr_2_fip(self):
+        if self.fip_ns.agent_gateway_port and not self.rtr_fip_connect:
+            ex_gw_port = self.get_ex_gw_port()
+            self.fip_ns.create_rtr_2_fip_link(self)
+            self.set_address_scope_interface_routes(ex_gw_port)
+            self.rtr_fip_connect = True
+            self.routes_updated([], self.router['routes'])
+
+    def _check_if_address_scopes_match(self, int_port, ex_gw_port):
+        """Checks and returns the matching state for v4 or v6 scopes."""
+        int_port_addr_scopes = int_port.get('address_scopes', {})
+        ext_port_addr_scopes = ex_gw_port.get('address_scopes', {})
+        key = (
+            lib_constants.IP_VERSION_6 if self._port_has_ipv6_subnet(int_port)
+            else lib_constants.IP_VERSION_4)
+        # NOTE: DVR does not support IPv6 for the floating namespace yet, so
+        # until we fix it, we probably should use the snat redirect path for
+        # the ports that have IPv6 address configured.
+        int_port_addr_value = int_port_addr_scopes.get(str(key))
+        # If the address scope of the interface is none, then don't need
+        # to compare and just return.
+        if int_port_addr_value is None:
+            return False
+        if ((key != lib_constants.IP_VERSION_6) and
+            int_port_addr_scopes.get(str(key)) in
+            ext_port_addr_scopes.values()):
+            return True
+        return False
+
+    def _delete_interface_route_in_fip_ns(self, router_port):
+        rtr_2_fip_ip, fip_2_rtr_name = self.get_rtr_fip_ip_and_interface_name()
+        fip_ns_name = self.fip_ns.get_name()
+        if ip_lib.network_namespace_exists(fip_ns_name):
+            device = ip_lib.IPDevice(fip_2_rtr_name, namespace=fip_ns_name)
+            if not device.exists():
+                return
+            for subnet in router_port['subnets']:
+                rtr_port_cidr = subnet['cidr']
+                device.route.delete_route(rtr_port_cidr, str(rtr_2_fip_ip))
+
+    def _add_interface_route_to_fip_ns(self, router_port):
+        rtr_2_fip_ip, fip_2_rtr_name = self.get_rtr_fip_ip_and_interface_name()
+        fip_ns_name = self.fip_ns.get_name()
+        if ip_lib.network_namespace_exists(fip_ns_name):
+            device = ip_lib.IPDevice(fip_2_rtr_name, namespace=fip_ns_name)
+            if not device.exists():
+                return
+            for subnet in router_port['subnets']:
+                rtr_port_cidr = subnet['cidr']
+                device.route.add_route(rtr_port_cidr, str(rtr_2_fip_ip))
+
+    def _add_interface_routing_rule_to_router_ns(self, router_port):
+        ip_rule = ip_lib.IPRule(namespace=self.ns_name)
+        for subnet in router_port['subnets']:
+            rtr_port_cidr = subnet['cidr']
+            ip_rule.rule.add(ip=rtr_port_cidr,
+                             table=dvr_fip_ns.FIP_RT_TBL,
+                             priority=dvr_fip_ns.FAST_PATH_EXIT_PR)
+
+    def _delete_interface_routing_rule_in_router_ns(self, router_port):
+        ip_rule = ip_lib.IPRule(namespace=self.ns_name)
+        for subnet in router_port['subnets']:
+            rtr_port_cidr = subnet['cidr']
+            ip_rule.rule.delete(ip=rtr_port_cidr,
+                                table=dvr_fip_ns.FIP_RT_TBL,
+                                priority=dvr_fip_ns.FAST_PATH_EXIT_PR)
+
+    def get_rtr_fip_ip_and_interface_name(self):
+        """Function that returns the router to fip interface name and ip."""
+        if self.rtr_fip_subnet is None:
+            self.rtr_fip_subnet = self.fip_ns.local_subnets.allocate(
+                self.router_id)
+        rtr_2_fip, __ = self.rtr_fip_subnet.get_pair()
+        fip_2_rtr_name = self.fip_ns.get_int_device_name(self.router_id)
+        return rtr_2_fip.ip, fip_2_rtr_name
+
+    def set_address_scope_interface_routes(self, ex_gw_port):
+        """Sets routing rules for router interfaces if addr scopes match."""
+        for port in self.internal_ports:
+            if self._check_if_address_scopes_match(port, ex_gw_port):
+                self._add_interface_routing_rule_to_router_ns(port)
+                self._add_interface_route_to_fip_ns(port)
+
+    def create_dvr_external_gateway_on_agent(self, ex_gw_port):
         fip_agent_port = self.get_floating_agent_gw_interface(
             ex_gw_port['network_id'])
-        if fip_agent_port:
+        if not fip_agent_port:
+            fip_agent_port = self.agent.plugin_rpc.get_agent_gateway_port(
+                self.agent.context, ex_gw_port['network_id'])
             LOG.debug("FloatingIP agent gateway port received from the "
-                "plugin: %s", fip_agent_port)
-        is_first = False
-        if floating_ips:
-            is_first = self.fip_ns.subscribe(ex_gw_port['network_id'])
-            if is_first and not fip_agent_port:
-                LOG.debug("No FloatingIP agent gateway port possibly due to "
-                          "late binding of the private port to the host, "
-                          "requesting agent gateway port for 'network-id' :"
-                          "%s", ex_gw_port['network_id'])
-                fip_agent_port = self.agent.plugin_rpc.get_agent_gateway_port(
-                    self.agent.context, ex_gw_port['network_id'])
-                if not fip_agent_port:
-                    LOG.error(_LE("No FloatingIP agent gateway port "
-                                  "returned from server for 'network-id': "
-                                  "%s"), ex_gw_port['network_id'])
-            if fip_agent_port:
-                if 'subnets' not in fip_agent_port:
-                    LOG.error(_LE('Missing subnet/agent_gateway_port'))
-                else:
-                    if is_first:
-                        self.fip_ns.create_gateway_port(fip_agent_port)
-                    else:
-                        self.fip_ns.update_gateway_port(fip_agent_port)
+                      "plugin: %s", fip_agent_port)
+        self.fip_ns.create_or_update_gateway_port(fip_agent_port)
 
-            if (self.fip_ns.agent_gateway_port and
-                (self.dist_fip_count == 0)):
-                self.fip_ns.create_rtr_2_fip_link(self)
+    def update_routing_table(self, operation, route):
+        # TODO(Swami): The static routes should be added to the
+        # specific namespace based on the availability of the
+        # network interfaces. In the case of DVR the static routes
+        # for local internal router networks can be added to the
+        # snat_namespace and router_namespace but should not be
+        # added to the fip namespace. Likewise the static routes
+        # for the external router networks should only be added to
+        # the snat_namespace and fip_namespace.
+        # The current code adds static routes to all namespaces in
+        # order to reduce the complexity. This should be revisited
+        # later.
+        if self.fip_ns and self.fip_ns.agent_gateway_port:
+            fip_ns_name = self.fip_ns.get_name()
+            agent_gw_port = self.fip_ns.agent_gateway_port
+            route_apply = self._check_if_route_applicable_to_fip_namespace(
+                route, agent_gw_port)
+            if route_apply:
+                if self.rtr_fip_subnet is None:
+                    self.rtr_fip_subnet = self.fip_ns.local_subnets.allocate(
+                        self.router_id)
+                rtr_2_fip, fip_2_rtr = self.rtr_fip_subnet.get_pair()
+                tbl_index = self._get_snat_idx(fip_2_rtr)
+                self._update_fip_route_table_with_next_hop_routes(
+                    operation, route, fip_ns_name, tbl_index)
+        super(DvrLocalRouter, self).update_routing_table(operation, route)
+
+    def _update_fip_route_table_with_next_hop_routes(
+        self, operation, route, fip_ns_name, tbl_index):
+        cmd = ['ip', 'route', operation, 'to', route['destination'],
+               'via', route['nexthop'], 'table', tbl_index]
+        ip_wrapper = ip_lib.IPWrapper(namespace=fip_ns_name)
+        if ip_wrapper.netns.exists(fip_ns_name):
+            ip_wrapper.netns.execute(cmd, check_exit_code=False)
+        else:
+            LOG.debug("The FIP namespace %(ns)s does not exist for "
+                      "router %(id)s",
+                      {'ns': fip_ns_name, 'id': self.router_id})
+
+    def _check_if_route_applicable_to_fip_namespace(
+        self, route, agent_gateway_port):
+        ip_cidrs = common_utils.fixed_ip_cidrs(agent_gateway_port['fixed_ips'])
+        nexthop_cidr = netaddr.IPAddress(route['nexthop'])
+        for gw_cidr in ip_cidrs:
+            gw_subnet_cidr = netaddr.IPNetwork(gw_cidr)
+            # NOTE: In the case of DVR routers apply the extra routes
+            # on the FIP namespace only if it is associated with the
+            # external agent gateway subnets.
+            if nexthop_cidr in gw_subnet_cidr:
+                return True
+        return False
 
     def get_router_cidrs(self, device):
         """As no floatingip will be set on the rfp device. Get floatingip from
@@ -571,10 +747,10 @@ class DvrLocalRouter(dvr_router_base.DvrRouterBase):
         return {common_utils.ip_to_cidr(route['cidr'])
                 for route in exist_routes}
 
-    def process(self, agent):
+    def process(self):
         ex_gw_port = self.get_ex_gw_port()
         if ex_gw_port:
-            self.fip_ns = agent.get_fip_ns(ex_gw_port['network_id'])
+            self.fip_ns = self.agent.get_fip_ns(ex_gw_port['network_id'])
             self.fip_ns.scan_fip_ports(self)
 
-        super(DvrLocalRouter, self).process(agent)
+        super(DvrLocalRouter, self).process()
