@@ -13,26 +13,25 @@
 #    under the License.
 
 import mock
-import netaddr
+from neutron_lib.api.definitions import portbindings
 from neutron_lib import constants as lib_constants
 from oslo_config import cfg
 from oslo_log import log
 from oslo_utils import uuidutils
 
-from neutron.agent.common import config as agent_config
 from neutron.agent.l3 import agent as l3_agent
+from neutron.agent.l3 import dvr_edge_router as dvr_edge_rtr
 from neutron.agent.l3 import dvr_local_router as dvr_router
-from neutron.agent.l3 import ha
 from neutron.agent.l3 import link_local_allocator as lla
 from neutron.agent.l3 import router_info
-from neutron.agent.linux import external_process
 from neutron.agent.linux import interface
 from neutron.agent.linux import ip_lib
 from neutron.common import constants as n_const
 from neutron.common import utils as common_utils
+from neutron.conf.agent import common as agent_config
 from neutron.conf.agent.l3 import config as l3_config
+from neutron.conf.agent.l3 import ha as ha_conf
 from neutron.conf import common as base_config
-from neutron.extensions import portbindings
 from neutron.tests import base
 from neutron.tests.common import l3_test_common
 
@@ -51,21 +50,21 @@ class TestDvrRouterOperations(base.BaseTestCase):
         log.register_options(self.conf)
         self.conf.register_opts(agent_config.AGENT_STATE_OPTS, 'AGENT')
         l3_config.register_l3_agent_config_opts(l3_config.OPTS, self.conf)
-        self.conf.register_opts(ha.OPTS)
+        ha_conf.register_l3_agent_ha_opts(self.conf)
         agent_config.register_interface_driver_opts_helper(self.conf)
         agent_config.register_process_monitor_opts(self.conf)
-        self.conf.register_opts(interface.OPTS)
-        self.conf.register_opts(external_process.OPTS)
+        agent_config.register_interface_opts(self.conf)
+        agent_config.register_external_process_opts(self.conf)
         self.conf.set_override('interface_driver',
                                'neutron.agent.linux.interface.NullDriver')
-        self.conf.set_override('send_arp_for_ha', 1)
         self.conf.set_override('state_path', cfg.CONF.state_path)
 
         self.device_exists_p = mock.patch(
             'neutron.agent.linux.ip_lib.device_exists')
         self.device_exists = self.device_exists_p.start()
 
-        self.ensure_dir = mock.patch('neutron.common.utils.ensure_dir').start()
+        self.ensure_dir = mock.patch(
+            'oslo_utils.fileutils.ensure_tree').start()
 
         mock.patch('neutron.agent.linux.keepalived.KeepalivedManager'
                    '.get_full_config_file_path').start()
@@ -75,7 +74,7 @@ class TestDvrRouterOperations(base.BaseTestCase):
         self.utils_exec = self.utils_exec_p.start()
 
         self.utils_replace_file_p = mock.patch(
-            'neutron.common.utils.replace_file')
+            'neutron_lib.utils.file.replace_file')
         self.utils_replace_file = self.utils_replace_file_p.start()
 
         self.external_process_p = mock.patch(
@@ -123,25 +122,25 @@ class TestDvrRouterOperations(base.BaseTestCase):
         self.snat_ports = [{'subnets': [{'cidr': '152.2.0.0/16',
                                          'gateway_ip': '152.2.0.1',
                                          'id': subnet_id_1}],
-                           'network_id': _uuid(),
-                           'device_owner':
-                           lib_constants.DEVICE_OWNER_ROUTER_SNAT,
-                           'mac_address': 'fa:16:3e:80:8d:80',
-                           'fixed_ips': [{'subnet_id': subnet_id_1,
-                                          'ip_address': '152.2.0.13',
-                                          'prefixlen': 16}],
-                           'id': _uuid(), 'device_id': _uuid()},
-                          {'subnets': [{'cidr': '152.10.0.0/16',
-                                        'gateway_ip': '152.10.0.1',
-                                        'id': subnet_id_2}],
-                           'network_id': _uuid(),
-                           'device_owner':
-                           lib_constants.DEVICE_OWNER_ROUTER_SNAT,
-                           'mac_address': 'fa:16:3e:80:8d:80',
-                           'fixed_ips': [{'subnet_id': subnet_id_2,
-                                         'ip_address': '152.10.0.13',
-                                         'prefixlen': 16}],
-                           'id': _uuid(), 'device_id': _uuid()}]
+                            'network_id': _uuid(),
+                            'device_owner':
+                            lib_constants.DEVICE_OWNER_ROUTER_SNAT,
+                            'mac_address': 'fa:16:3e:80:8d:80',
+                            'fixed_ips': [{'subnet_id': subnet_id_1,
+                                           'ip_address': '152.2.0.13',
+                                           'prefixlen': 16}],
+                            'id': _uuid(), 'device_id': _uuid()},
+                           {'subnets': [{'cidr': '152.10.0.0/16',
+                                         'gateway_ip': '152.10.0.1',
+                                         'id': subnet_id_2}],
+                            'network_id': _uuid(),
+                            'device_owner':
+                            lib_constants.DEVICE_OWNER_ROUTER_SNAT,
+                            'mac_address': 'fa:16:3e:80:8d:80',
+                            'fixed_ips': [{'subnet_id': subnet_id_2,
+                                           'ip_address': '152.10.0.13',
+                                           'prefixlen': 16}],
+                            'id': _uuid(), 'device_id': _uuid()}]
 
         self.ri_kwargs = {'agent_conf': self.conf,
                           'interface_driver': self.mock_driver}
@@ -151,13 +150,21 @@ class TestDvrRouterOperations(base.BaseTestCase):
         self.router_id = _uuid()
         if not router:
             router = mock.MagicMock()
-        return dvr_router.DvrLocalRouter(agent,
-                                    HOSTNAME,
-                                    self.router_id,
-                                    router,
-                                    self.conf,
-                                    mock.Mock(),
-                                    **kwargs)
+        kwargs['agent'] = agent
+        kwargs['router_id'] = self.router_id
+        kwargs['router'] = router
+        kwargs['agent_conf'] = self.conf
+        kwargs['interface_driver'] = mock.Mock()
+        return dvr_router.DvrLocalRouter(HOSTNAME, **kwargs)
+
+    def _set_ri_kwargs(self, agent, router_id, router):
+        self.ri_kwargs['agent'] = agent
+        self.ri_kwargs['router_id'] = router_id
+        self.ri_kwargs['router'] = router
+
+    def test_gw_ns_name(self):
+        ri = self._create_router()
+        self.assertEqual(ri.ns_name, ri.get_gw_ns_name())
 
     def test_create_dvr_fip_interfaces_update(self):
         ri = self._create_router()
@@ -167,11 +174,54 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.get_floating_ips = mock.Mock(return_value=True)
         ri.fip_ns = mock.Mock()
         ri.fip_ns.subscribe.return_value = False
+        ri.rtr_fip_connect = True
         ex_gw_port = {'network_id': 'fake_net_id'}
-        ri.create_dvr_fip_interfaces(ex_gw_port)
-        ri.fip_ns.update_gateway_port.assert_called_once_with(
+        ri.create_dvr_external_gateway_on_agent(ex_gw_port)
+        ri.fip_ns.create_or_update_gateway_port.assert_called_once_with(
             fip_agent_port)
 
+    def test_create_dvr_fip_interfaces_with_matching_address_scope(self):
+        self._setup_create_dvr_fip_interfaces_for_setting_routing_rules(
+            address_scopes_match=True)
+
+    def test_create_dvr_fip_interfaces_with_address_scope_mismatch(self):
+        self._setup_create_dvr_fip_interfaces_for_setting_routing_rules()
+
+    def _setup_create_dvr_fip_interfaces_for_setting_routing_rules(
+        self, address_scopes_match=False):
+        ri = self._create_router()
+        ri.get_floating_agent_gw_interface = mock.Mock()
+        ri.fip_ns = mock.Mock()
+        ri._add_interface_routing_rule_to_router_ns = mock.Mock()
+        ri._add_interface_route_to_fip_ns = mock.Mock()
+        ri.fip_ns._create_rtr_2_fip_link = mock.Mock()
+        ri.internal_ports = ['moke_port_1', 'moke_port_2']
+        if address_scopes_match:
+            ri._check_if_address_scopes_match = mock.Mock(
+                return_value=True)
+        else:
+            ri._check_if_address_scopes_match = mock.Mock(
+                return_value=False)
+        ri.rtr_fip_connect = False
+        ex_gw_port = {'network_id': 'fake_net_id'}
+        ri.create_dvr_external_gateway_on_agent(ex_gw_port)
+        ri.connect_rtr_2_fip()
+        self.assertTrue(ri._check_if_address_scopes_match.called)
+        if address_scopes_match:
+            self.assertTrue(
+                ri.fip_ns.create_rtr_2_fip_link.called)
+            self.assertTrue(
+                ri._add_interface_routing_rule_to_router_ns.called)
+            self.assertTrue(
+                ri._add_interface_route_to_fip_ns.called)
+        else:
+            self.assertFalse(
+                ri._add_interface_routing_rule_to_router_ns.called)
+            self.assertFalse(
+                ri._add_interface_route_to_fip_ns.called)
+            self.assertTrue(
+                ri.fip_ns.create_rtr_2_fip_link.called)
+
     def test_get_floating_ips_dvr(self):
         router = mock.MagicMock()
         router.get.return_value = [{'host': HOSTNAME},
@@ -180,7 +230,75 @@ class TestDvrRouterOperations(base.BaseTestCase):
 
         fips = ri.get_floating_ips()
 
-        self.assertEqual([{'host': HOSTNAME}], fips)
+        self.assertEqual(
+            [{'host': HOSTNAME}, {'host': mock.sentinel.otherhost}], fips)
+
+    def test_floating_forward_rules_no_fip_ns(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        fip = {'id': _uuid()}
+        ri = self._create_router(router)
+        self.assertFalse(ri.floating_forward_rules(fip))
+
+    def test_floating_forward_rules(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        ri = self._create_router(router)
+        floating_ip = '15.1.2.3'
+        rtr_2_fip_name = 'fake_router'
+        fixed_ip = '192.168.0.1'
+        fip = {'id': _uuid(),
+               'fixed_ip_address': '192.168.0.1',
+               'floating_ip_address': '15.1.2.3'}
+        instance = mock.Mock()
+        instance.get_rtr_ext_device_name = mock.Mock(
+                                               return_value=rtr_2_fip_name)
+        ri.fip_ns = instance
+        dnat_from_floatingip_to_fixedip = (
+            'PREROUTING', '-d %s/32 -i %s -j DNAT --to-destination %s' % (
+                floating_ip, rtr_2_fip_name, fixed_ip))
+        snat_from_fixedip_to_floatingip = (
+            'float-snat', '-s %s/32 -j SNAT --to-source %s' % (
+                fixed_ip, floating_ip))
+        actual = ri.floating_forward_rules(fip)
+        expected = [dnat_from_floatingip_to_fixedip,
+                    snat_from_fixedip_to_floatingip]
+        self.assertEqual(expected, actual)
+
+    def test_floating_mangle_rules_no_fip_ns(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        ri = self._create_router(router)
+        floating_ip = mock.Mock()
+        fixed_ip = mock.Mock()
+        internal_mark = mock.Mock()
+        self.assertFalse(ri.floating_mangle_rules(floating_ip, fixed_ip,
+                                                  internal_mark))
+
+    def test_floating_mangle_rules(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        ri = self._create_router(router)
+        floating_ip = '15.1.2.3'
+        fixed_ip = '192.168.0.1'
+        internal_mark = 'fake_mark'
+        rtr_2_fip_name = 'fake_router'
+        instance = mock.Mock()
+        instance.get_rtr_ext_device_name = mock.Mock(
+                                               return_value=rtr_2_fip_name)
+        ri.fip_ns = instance
+        mark_traffic_to_floating_ip = (
+            'floatingip', '-d %s/32 -i %s -j MARK --set-xmark %s' % (
+                floating_ip, rtr_2_fip_name, internal_mark))
+        mark_traffic_from_fixed_ip = (
+            'FORWARD', '-s %s/32 -j $float-snat' % fixed_ip)
+        actual = ri.floating_mangle_rules(floating_ip, fixed_ip, internal_mark)
+        expected = [mark_traffic_to_floating_ip, mark_traffic_from_fixed_ip]
+        self.assertEqual(expected, actual)
 
     @mock.patch.object(ip_lib, 'send_ip_addr_adv_notif')
     @mock.patch.object(ip_lib, 'IPDevice')
@@ -188,6 +306,7 @@ class TestDvrRouterOperations(base.BaseTestCase):
     def test_floating_ip_added_dist(self, mIPRule, mIPDevice, mock_adv_notif):
         router = mock.MagicMock()
         ri = self._create_router(router)
+        ri.ex_gw_port = ri.router['gw_port']
         ext_net_id = _uuid()
         subnet_id = _uuid()
         agent_gw_port = {'fixed_ips': [{'ip_address': '20.0.0.30',
@@ -208,22 +327,23 @@ class TestDvrRouterOperations(base.BaseTestCase):
                'port_id': _uuid()}
         ri.fip_ns = mock.Mock()
         ri.fip_ns.agent_gateway_port = agent_gw_port
+        ri.create_dvr_external_gateway_on_agent(ri.ex_gw_port)
+        ri.connect_rtr_2_fip()
+        self.assertTrue(ri.rtr_fip_connect)
         ri.fip_ns.allocate_rule_priority.return_value = FIP_PRI
         subnet = lla.LinkLocalAddressPair('169.254.30.42/31')
         ri.rtr_fip_subnet = subnet
         ri.fip_ns.local_subnets = mock.Mock()
         ri.fip_ns.local_subnets.allocate.return_value = subnet
-        ri.dist_fip_count = 0
         ip_cidr = common_utils.ip_to_cidr(fip['floating_ip_address'])
         ri.floating_ip_added_dist(fip, ip_cidr)
         mIPRule().rule.add.assert_called_with(ip='192.168.0.1',
                                               table=16,
                                               priority=FIP_PRI)
         ri.fip_ns.local_subnets.allocate.assert_not_called()
-        self.assertEqual(1, ri.dist_fip_count)
 
         # Validate that fip_ns.local_subnets is called when
-        # rtr_fip_subnet is None
+        # ri.rtr_fip_subnet is None
         ri.rtr_fip_subnet = None
         ri.floating_ip_added_dist(fip, ip_cidr)
         mIPRule().rule.add.assert_called_with(ip='192.168.0.1',
@@ -238,9 +358,10 @@ class TestDvrRouterOperations(base.BaseTestCase):
     def test_floating_ip_removed_dist(self, mIPRule, mIPDevice, mIPWrapper):
         router = mock.MagicMock()
         ri = self._create_router(router)
-
+        ri.ex_gw_port = ri.router['gw_port']
         subnet_id = _uuid()
-        agent_gw_port = {'fixed_ips': [{'ip_address': '20.0.0.30',
+        fixed_ip = '20.0.0.30'
+        agent_gw_port = {'fixed_ips': [{'ip_address': fixed_ip,
                                         'prefixlen': 24,
                                         'subnet_id': subnet_id}],
                          'subnets': [{'id': subnet_id,
@@ -250,11 +371,9 @@ class TestDvrRouterOperations(base.BaseTestCase):
                          'network_id': _uuid(),
                          'mac_address': 'ca:fe:de:ad:be:ef'}
         fip_cidr = '11.22.33.44/24'
-
-        ri.dist_fip_count = 2
         ri.fip_ns = mock.Mock()
         ri.fip_ns.get_name.return_value = 'fip_ns_name'
-        ri.floating_ips_dict['11.22.33.44'] = FIP_PRI
+        ri.floating_ips_dict['11.22.33.44'] = (fixed_ip, FIP_PRI)
         ri.fip_2_rtr = '11.22.33.42'
         ri.rtr_2_fip = '11.22.33.40'
         ri.fip_ns.agent_gateway_port = agent_gw_port
@@ -263,50 +382,40 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.fip_ns.local_subnets = mock.Mock()
         ri.floating_ip_removed_dist(fip_cidr)
         mIPRule().rule.delete.assert_called_with(
-            ip=str(netaddr.IPNetwork(fip_cidr).ip), table=16, priority=FIP_PRI)
+            ip=fixed_ip, table=16, priority=FIP_PRI)
         mIPDevice().route.delete_route.assert_called_with(fip_cidr, str(s.ip))
         ri.fip_ns.local_subnets.allocate.assert_not_called()
 
-        ri.dist_fip_count = 1
-        s1 = lla.LinkLocalAddressPair('15.1.2.3/32')
-        ri.rtr_fip_subnet = None
-        ri.fip_ns.local_subnets.allocate.return_value = s1
-        _, fip_to_rtr = s1.get_pair()
-        fip_ns = ri.fip_ns
-        ri.floating_ip_removed_dist(fip_cidr)
-        self.assertTrue(fip_ns.destroyed)
-        mIPWrapper().del_veth.assert_called_once_with(
-            fip_ns.get_int_device_name(router['id']))
-        mIPDevice().route.delete_gateway.assert_called_once_with(
-            str(fip_to_rtr.ip), table=16)
-        self.assertFalse(ri.fip_ns.unsubscribe.called)
-        ri.fip_ns.local_subnets.allocate.assert_called_once_with(ri.router_id)
-
     @mock.patch.object(ip_lib, 'IPRule')
     def test_floating_ip_moved_dist(self, mIPRule):
         router = mock.MagicMock()
         ri = self._create_router(router)
         floating_ip_address = '15.1.2.3'
+        fixed_ip = '192.168.0.1'
         fip = {'floating_ip_address': floating_ip_address,
-               'fixed_ip_address': '192.168.0.1'}
-        ri.floating_ips_dict['15.1.2.3'] = FIP_PRI
+               'fixed_ip_address': fixed_ip}
+        ri.floating_ips_dict['15.1.2.3'] = (fixed_ip, FIP_PRI)
         ri.fip_ns = mock.Mock()
         ri.fip_ns.allocate_rule_priority.return_value = FIP_PRI
         ri.floating_ip_moved_dist(fip)
 
         mIPRule().rule.delete.assert_called_once_with(
-            ip=floating_ip_address, table=16, priority=FIP_PRI)
+            ip=fixed_ip, table=16, priority=FIP_PRI)
         ri.fip_ns.deallocate_rule_priority.assert_called_once_with(
             floating_ip_address)
         ri.fip_ns.allocate_rule_priority.assert_called_once_with(
             floating_ip_address)
-        mIPRule().rule.add.assert_called_with(ip='192.168.0.1',
+        mIPRule().rule.add.assert_called_with(ip=fixed_ip,
                                               table=16,
                                               priority=FIP_PRI)
 
-    def _test_add_floating_ip(self, ri, fip, is_failure):
-        ri.floating_ip_added_dist = mock.Mock()
-
+    def _test_add_floating_ip(self, ri, fip, is_failure=False):
+        if not is_failure:
+            ri.floating_ip_added_dist = mock.Mock(
+                return_value=lib_constants.FLOATINGIP_STATUS_ACTIVE)
+        else:
+            ri.floating_ip_added_dist = mock.Mock(
+                return_value=lib_constants.FLOATINGIP_STATUS_ERROR)
         result = ri.add_floating_ip(fip,
                                     mock.sentinel.interface_name,
                                     mock.sentinel.device)
@@ -318,10 +427,18 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri = self._create_router(mock.MagicMock())
         ip = '15.1.2.3'
         fip = {'floating_ip_address': ip}
-        result = self._test_add_floating_ip(ri, fip, True)
+        result = self._test_add_floating_ip(ri, fip)
         ri.floating_ip_added_dist.assert_called_once_with(fip, ip + '/32')
         self.assertEqual(lib_constants.FLOATINGIP_STATUS_ACTIVE, result)
 
+    def test_add_floating_ip_failure(self):
+        ri = self._create_router(mock.MagicMock())
+        ip = '15.1.2.3'
+        fip = {'floating_ip_address': ip}
+        result = self._test_add_floating_ip(ri, fip, True)
+        ri.floating_ip_added_dist.assert_called_once_with(fip, ip + '/32')
+        self.assertEqual(lib_constants.FLOATINGIP_STATUS_ERROR, result)
+
     @mock.patch.object(router_info.RouterInfo, 'remove_floating_ip')
     def test_remove_floating_ip(self, super_remove_floating_ip):
         ri = self._create_router(mock.MagicMock())
@@ -373,15 +490,26 @@ class TestDvrRouterOperations(base.BaseTestCase):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
         router = l3_test_common.prepare_router_data(num_internal_ports=2)
         router['distributed'] = True
-        ri = dvr_router.DvrLocalRouter(
-            agent, HOSTNAME, router['id'], router, **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
         ports = ri.router.get(lib_constants.INTERFACE_KEY, [])
         subnet_id = l3_test_common.get_subnet_id(ports[0])
         test_ports = [{'mac_address': '00:11:22:33:44:55',
-                      'device_owner': lib_constants.DEVICE_OWNER_DHCP,
-                      'fixed_ips': [{'ip_address': '1.2.3.4',
-                                     'prefixlen': 24,
-                                     'subnet_id': subnet_id}]}]
+                       'device_owner': lib_constants.DEVICE_OWNER_DHCP,
+                       'fixed_ips': [{'ip_address': '1.2.3.4',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]},
+                      {'mac_address': '11:22:33:44:55:66',
+                       'device_owner': lib_constants.DEVICE_OWNER_LOADBALANCER,
+                       'fixed_ips': [{'ip_address': '1.2.3.5',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]},
+                      {'mac_address': '22:33:44:55:66:77',
+                       'device_owner':
+                           lib_constants.DEVICE_OWNER_LOADBALANCERV2,
+                       'fixed_ips': [{'ip_address': '1.2.3.6',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]}]
 
         self.plugin_api.get_ports_by_subnet.return_value = test_ports
 
@@ -393,7 +521,49 @@ class TestDvrRouterOperations(base.BaseTestCase):
             ri._set_subnet_arp_info(subnet_id)
         self.assertEqual(1, parp.call_count)
         self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.2.3.4', '00:11:22:33:44:55')
+            '1.2.3.4', '00:11:22:33:44:55', nud_state='permanent')
+
+        # Test negative case
+        router['distributed'] = False
+        ri._set_subnet_arp_info(subnet_id)
+        self.mock_ip_dev.neigh.add.never_called()
+
+    def test__set_subnet_arp_info_with_allowed_address_pair_port(self):
+        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
+        router = l3_test_common.prepare_router_data(num_internal_ports=2)
+        router['distributed'] = True
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
+        ports = ri.router.get(lib_constants.INTERFACE_KEY, [])
+        subnet_id = l3_test_common.get_subnet_id(ports[0])
+        test_ports = [{'mac_address': '00:11:22:33:44:55',
+                       'device_owner': '',
+                       'fixed_ips': [{'ip_address': '1.2.3.4',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]},
+                      {'mac_address': '11:22:33:44:55:66',
+                       'device_owner': lib_constants.DEVICE_OWNER_LOADBALANCER,
+                       'fixed_ips': [{'ip_address': '1.2.3.5',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]},
+                      {'mac_address': '22:33:44:55:66:77',
+                       'device_owner':
+                           lib_constants.DEVICE_OWNER_LOADBALANCERV2,
+                       'fixed_ips': [{'ip_address': '1.2.3.6',
+                                      'prefixlen': 24,
+                                      'subnet_id': subnet_id}]}]
+
+        self.plugin_api.get_ports_by_subnet.return_value = test_ports
+
+        # Test basic case
+        ports[0]['subnets'] = [{'id': subnet_id,
+                                'cidr': '1.2.3.0/24'}]
+        with mock.patch.object(ri,
+                               '_process_arp_cache_for_internal_port') as parp:
+            ri._set_subnet_arp_info(subnet_id)
+        self.assertEqual(1, parp.call_count)
+        self.mock_ip_dev.neigh.add.assert_called_once_with(
+            '1.2.3.4', '00:11:22:33:44:55', nud_state='reachable')
 
         # Test negative case
         router['distributed'] = False
@@ -408,14 +578,33 @@ class TestDvrRouterOperations(base.BaseTestCase):
             router[lib_constants.INTERFACE_KEY][0])
         arp_table = {'ip_address': '1.7.23.11',
                      'mac_address': '00:11:22:33:44:55',
-                     'subnet_id': subnet_id}
+                     'subnet_id': subnet_id,
+                     'nud_state': 'permanent'}
 
         payload = {'arp_table': arp_table, 'router_id': router['id']}
         agent._router_added(router['id'], router)
         agent.add_arp_entry(None, payload)
         agent.router_deleted(None, router['id'])
         self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.7.23.11', '00:11:22:33:44:55')
+            '1.7.23.11', '00:11:22:33:44:55', nud_state='permanent')
+
+    def test_add_arp_entry_with_nud_state_reachable(self):
+        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
+        router = l3_test_common.prepare_router_data(num_internal_ports=2)
+        router['distributed'] = True
+        subnet_id = l3_test_common.get_subnet_id(
+            router[lib_constants.INTERFACE_KEY][0])
+        arp_table = {'ip_address': '1.7.23.11',
+                     'mac_address': '00:11:22:33:44:55',
+                     'subnet_id': subnet_id,
+                     'nud_state': 'reachable'}
+
+        payload = {'arp_table': arp_table, 'router_id': router['id']}
+        agent._router_added(router['id'], router)
+        agent.add_arp_entry(None, payload)
+        agent.router_deleted(None, router['id'])
+        self.mock_ip_dev.neigh.add.assert_called_once_with(
+            '1.7.23.11', '00:11:22:33:44:55', nud_state='reachable')
 
     def test_add_arp_entry_no_routerinfo(self):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
@@ -430,12 +619,10 @@ class TestDvrRouterOperations(base.BaseTestCase):
         agent.add_arp_entry(None, payload)
 
     def test__update_arp_entry_with_no_subnet(self):
-        ri = dvr_router.DvrLocalRouter(
-            mock.sentinel.agent,
-            HOSTNAME,
-            'foo_router_id',
-            {'distributed': True, 'gw_port_host': HOSTNAME},
-            **self.ri_kwargs)
+        self._set_ri_kwargs(mock.sentinel.agent,
+                            'foo_router_id',
+                            {'distributed': True, 'gw_port_host': HOSTNAME})
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
         with mock.patch.object(l3_agent.ip_lib, 'IPDevice') as f:
             ri._update_arp_entry(mock.ANY, mock.ANY, 'foo_subnet_id', 'add')
         self.assertFalse(f.call_count)
@@ -444,8 +631,8 @@ class TestDvrRouterOperations(base.BaseTestCase):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
         router = l3_test_common.prepare_router_data(num_internal_ports=2)
         router['distributed'] = True
-        ri = dvr_router.DvrLocalRouter(
-            agent, HOSTNAME, router['id'], router, **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
         subnet_id = l3_test_common.get_subnet_id(
             ri.router[lib_constants.INTERFACE_KEY][0])
         return ri, subnet_id
@@ -523,8 +710,8 @@ class TestDvrRouterOperations(base.BaseTestCase):
         router[n_const.FLOATINGIP_AGENT_INTF_KEY] = agent_gateway_port
         router['distributed'] = True
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        ri = dvr_router.DvrLocalRouter(
-            agent, HOSTNAME, router['id'], router, **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
         self.assertEqual(
             agent_gateway_port[0],
             ri.get_floating_agent_gw_interface(fake_network_id))
@@ -548,13 +735,9 @@ class TestDvrRouterOperations(base.BaseTestCase):
         router[lib_constants.FLOATINGIP_KEY] = fake_floatingips['floatingips']
         router['distributed'] = True
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        ri = dvr_router.DvrLocalRouter(agent,
-                                  HOSTNAME,
-                                  router['id'],
-                                  router,
-                                  **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
         ri.iptables_manager.ipv4['nat'] = mock.MagicMock()
-        ri.dist_fip_count = 0
         fip_ns = agent.get_fip_ns(mock.sentinel.ext_net_id)
         subnet_id = _uuid()
         fip_ns.agent_gateway_port = (
@@ -572,11 +755,8 @@ class TestDvrRouterOperations(base.BaseTestCase):
                                             agent_mode, expected_call_count):
         router = l3_test_common.prepare_router_data(num_internal_ports=2)
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        ri = dvr_router.DvrLocalRouter(agent,
-                                       HOSTNAME,
-                                       router['id'],
-                                       router,
-                                       **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
 
         interface_name, ex_gw_port = l3_test_common.prepare_ext_gw_test(self,
                                                                         ri)
@@ -602,14 +782,14 @@ class TestDvrRouterOperations(base.BaseTestCase):
 
     def test_external_gateway_removed_ext_gw_port_and_fip(self):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        agent.conf.agent_mode = 'dvr'
+        agent.conf.agent_mode = lib_constants.L3_AGENT_MODE_DVR_SNAT
         router = l3_test_common.prepare_router_data(num_internal_ports=2)
         router['gw_port_host'] = HOSTNAME
         self.mock_driver.unplug.reset_mock()
 
         external_net_id = router['gw_port']['network_id']
-        ri = dvr_router.DvrLocalRouter(
-            agent, HOSTNAME, router['id'], router, **self.ri_kwargs)
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_edge_rtr.DvrEdgeRouter(HOSTNAME, **self.ri_kwargs)
         ri.remove_floating_ip = mock.Mock()
         agent._fetch_external_net_id = mock.Mock(return_value=external_net_id)
         ri.ex_gw_port = ri.router['gw_port']
@@ -636,7 +816,6 @@ class TestDvrRouterOperations(base.BaseTestCase):
 
         vm_floating_ip = '19.4.4.2'
         ri.floating_ips_dict[vm_floating_ip] = FIP_PRI
-        ri.dist_fip_count = 1
         ri.rtr_fip_subnet = ri.fip_ns.local_subnets.allocate(ri.router_id)
         _, fip_to_rtr = ri.rtr_fip_subnet.get_pair()
         self.mock_ip.get_devices.return_value = [
@@ -644,10 +823,49 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.get_router_cidrs = mock.Mock(
             return_value={vm_floating_ip + '/32', '19.4.4.1/24'})
         self.device_exists.return_value = True
-
         ri.external_gateway_removed(
             ri.ex_gw_port,
             ri.get_external_device_name(ri.ex_gw_port['id']))
-
         ri.remove_floating_ip.assert_called_once_with(self.mock_ip_dev,
                                                       '19.4.4.2/32')
+
+    def test_get_router_cidrs_no_fip_ns(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        ri = self._create_router(router)
+        device = mock.Mock()
+        self.assertFalse(ri.get_router_cidrs(device))
+
+    def test_get_router_cidrs_no_device_exists(self):
+        router = mock.MagicMock()
+        router.get.return_value = [{'host': HOSTNAME},
+                                   {'host': mock.sentinel.otherhost}]
+        ri = self._create_router(router)
+        fake_fip_ns = mock.Mock(return_value=True)
+        fake_fip_ns.get_name = mock.Mock(return_value=None)
+        fake_fip_ns.get_int_device_name = mock.Mock(return_value=None)
+        ri.fip_ns = fake_fip_ns
+        device = mock.Mock()
+        device.exists = mock.Mock(return_value=False)
+        with mock.patch.object(ip_lib, 'IPDevice', return_value=device):
+            self.assertFalse(ri.get_router_cidrs(device))
+
+    @mock.patch.object(router_info.RouterInfo, '_add_snat_rules')
+    @mock.patch.object(dvr_router.DvrLocalRouter, '_handle_router_snat_rules')
+    def test_handle_snat_rule_for_centralized_fip(
+            self, _add_snat_rules, _handle_router_snat_rules):
+        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
+        agent.conf.agent_mode = lib_constants.L3_AGENT_MODE_DVR_SNAT
+        self.mock_driver.unplug.reset_mock()
+
+        router = l3_test_common.prepare_router_data(enable_floating_ip=True)
+        router['gw_port_host'] = HOSTNAME
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri = dvr_edge_rtr.DvrEdgeRouter(HOSTNAME, **self.ri_kwargs)
+        ri.snat_iptables_manager = mock.MagicMock()
+        ipv4_nat = ri.snat_iptables_manager.ipv4['nat']
+        interface_name, ex_gw_port = l3_test_common.prepare_ext_gw_test(self,
+                                                                        ri)
+        ri._handle_router_snat_rules(ex_gw_port, interface_name)
+        ipv4_nat.add_rule.assert_called_once_with('snat', '-j $float-snat')
